/*
 * Tcl-Extension for access to the eggmcs913-driver
 *
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <tcl.h>
#include <blt.h>

#include "tclMCS.h"
#include <eggmcs913.h>


static void MCSDelete(ClientData clientData)
{
	MCS *mcsPtr = (MCS *) clientData;

	close(mcsPtr->mcsfd);

	if (mcsPtr->data != NULL) {
		ckfree(mcsPtr->data);
	}

	ckfree(mcsPtr);
}

static int
MCSObjectCmd(ClientData clientData, Tcl_Interp* interp, int argc, char** argv)
{
	int result = TCL_OK;
	int length;
	char c;
	int stat;

	MCS *mcsPtr = (MCS *) clientData;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " option ?arg arg ...?\"",
				 (char *) NULL);
		return TCL_ERROR;
	}

	c = argv[1][0];
	length = strlen(argv[1]);

	if ((c == 'd') && (strncmp(argv[1], "data", length) == 0)) {
		int i;
		int start, end;
		int bytecount, cnt, channel_count;

		char* yvec = NULL;


		if (argc != 5) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " data start end yvec\"",
					 (char *) NULL);
			goto error;
		}

		if ( Tcl_GetInt(interp, argv[2], &start) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for start");
			return TCL_ERROR;
		}
		if ( Tcl_GetInt(interp, argv[3], &end) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for end");
			return TCL_ERROR;
		}
		yvec = argv[4];

		if (start < 0) {
			sprintf(interp->result, "start < 0");
			return TCL_ERROR;
		}
		if (end < 0) {
			sprintf(interp->result, "end < 0");
			return TCL_ERROR;
		}
		if (end >= mcsPtr->nchannels) end = mcsPtr->nchannels-1;
		if (end<start) {
			sprintf(interp->result, "end < start");
			return TCL_ERROR;
		}


		if ( (stat = ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_STARTCHANNEL,
				   start )) != 0) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_IOCTL_STARTCHANNEL failed\n" );
			return TCL_ERROR;
		}

		channel_count = end - start + 1;
		cnt = channel_count; 
		bytecount = cnt * sizeof(unsigned long);

		if (mcsPtr->dsize < bytecount) {
			if (mcsPtr->data != NULL) {
				ckfree(mcsPtr->data);
			}
			mcsPtr->data = (unsigned long *) ckalloc(bytecount);
			if (mcsPtr->data == NULL) {
				sprintf(interp->result, "no memory for data");
				return TCL_ERROR;
			}
           
			mcsPtr->dsize = bytecount;
		}

		stat = read(mcsPtr->mcsfd, mcsPtr->data, bytecount);
		if (stat != bytecount) {
			sprintf(interp->result, "read data failed");
			return TCL_ERROR;
		}

		/* resize vector if necessary */
		if (channel_count > mcsPtr->ysize) {
			if (mcsPtr->ydata.valueArr != NULL) {
				ckfree(mcsPtr->ydata.valueArr);
			}
			mcsPtr->ydata.valueArr = (double*)
				ckalloc(channel_count*sizeof(double));
			if (mcsPtr->ydata.valueArr == NULL) {
				sprintf(interp->result, "no memory for ydata");
				return TCL_ERROR;
			}
			mcsPtr->ysize = channel_count;
			mcsPtr->ydata.arraySize = channel_count;
			mcsPtr->ydata.numValues = channel_count;
		}

		for (i=0; i<channel_count; i++) {
			mcsPtr->ydata.valueArr[i] = mcsPtr->data[i];
		}

		mcsPtr->ydata.numValues = channel_count;
		if (Blt_ResetVector(interp, yvec, &mcsPtr->ydata,
				    TCL_STATIC) != TCL_OK) {
			sprintf(interp->result, "y-ResetVector failed\n");
			return TCL_ERROR;
		}

	} else if ((c == 's') && (strncmp(argv[1], "start", length) == 0)) {
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " start\"", (char *) NULL);
			goto error;
		}

		stat = ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_START, -1);
		if (stat != 0 ) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_IOCTL_START failed\n" );
			goto error;
		}
	} else if ((c == 's') && (strncmp(argv[1], "stop", length) == 0)) {
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " stop\"", (char *) NULL);
			goto error;
		}

		stat = ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_STOP, -1);
		if (stat != 0 ) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_IOCTL_STOP failed\n" );
			goto error;
		}
	} else if ((c == 'r') &&
		   (strncmp(argv[1], "requeststop", length) == 0)) {
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " requeststop\"",
					 (char *) NULL);
			goto error;
		}

		stat = ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_STOPREQ, -1);
		if (stat != 0) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_IOCTL_STOPREQ failed\n" );
			goto error;
		}
	} else if ((c == 'f') && (strncmp(argv[1], "fd", length) == 0)) {
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " fd\"", (char *) NULL);
			goto error;
		}
		sprintf(interp->result, "%d", mcsPtr->mcsfd);
	} else if ((c == 'c') && (strncmp(argv[1], "clear", length) == 0)) {
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " clear\"", (char *) NULL);
			goto error;
		}

		stat = ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_CLEAR, -1);
		if (stat != 0) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_IOCTL_CLEAR failed\n" );
			goto error;
		}
	} else if ((c == 'a') && (strncmp(argv[1], "active", length) == 0)) {
		int active;
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " active\"", (char *) NULL);
			goto error;
		}

		stat = ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_ACTIVE, &active);
		if (stat != 0) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_IOCTL_ACTIVE failed\n" );
			goto error;
		}
		sprintf(interp->result, "%d", active);
	} else if ((c == 'd') && (strncmp(argv[1],"dwelltime", length) == 0)) {
		unsigned int dt;
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " dwelltime\"",
					 (char *) NULL);
			goto error;
		}

		stat =ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_GET_DWELLTIME, &dt);
		if (stat != 0) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_GET_DWELLTIME failed\n" );
			goto error;
		}
		sprintf(interp->result, "%d", dt);

	} else if ((c == 'p') && 
		   (strncmp(argv[1], "passcount", length) == 0)) {
		unsigned int pc;
		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " passcount\"",
					 (char *) NULL);
			goto error;
		}

		stat=ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_GET_PASSCOUNT, &pc);
		if (stat != 0) {
			sprintf(interp->result, "ioctl: "
				"EGGMCS913_GET_PASSCOUNT failed\n" );
			goto error;
		}
		sprintf(interp->result, "%d", pc);

	} else if ((c == 'p') && (strncmp(argv[1], "params", length) == 0)) {
		int dt, pl, pcp, extst, extdt, add;

		if (argc != 8) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0],
					 " dwelltime passlength "
					 "passcountpreset extstart "
					 "extdwelltime sum\"",
					 (char *) NULL);
			goto error;
		}

		if ( Tcl_GetInt(interp, argv[2], &dt) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for "
				"dwelltime");
			return TCL_ERROR;
		}
		if ( Tcl_GetInt(interp, argv[3], &pl) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for "
				"passlength");
			return TCL_ERROR;
		}
		if ( Tcl_GetInt(interp, argv[4], &pcp) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for "
				"passcountpreset");
			return TCL_ERROR;
		}
		if ( Tcl_GetBoolean(interp, argv[5], &extst) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for extstart");
			return TCL_ERROR;
		}
		if ( Tcl_GetBoolean(interp, argv[6], &extdt) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for "
				"extdwelltime");
			return TCL_ERROR;
		}
		if ( Tcl_GetBoolean(interp, argv[7], &add) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for sum");
			return TCL_ERROR;
		}

		/* Configure the board */
		if ( ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_SET_DWELLTIME, dt)
		     != 0) {
			sprintf(interp->result, "failed to set dwelltime %d",
				dt);
			return TCL_ERROR;
		}
		if ( ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_SET_PASSLENGTH, pl)
		     != 0) {
			sprintf(interp->result, "failed to set passlength %d",
				pl);
			return TCL_ERROR;
		}
		if ( ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_SET_PASSCOUNTPRESET,
			   pcp) != 0) {
			sprintf(interp->result,
				"failed to set passcountpreset %d", pcp);
			return TCL_ERROR;
		}
		if ( ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_SET_EXTSTART, extst)
		     != 0) {
			sprintf(interp->result, "failed to set extst %d",
				extst);
			return TCL_ERROR;
		}
		if ( ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_SET_EXTDWELLTIME,
			   extdt) != 0) {
			sprintf(interp->result,
				"failed to set extdt %d", extdt);
			return TCL_ERROR;
		}
		if ( ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_SET_ADD, add) != 0) {
			sprintf(interp->result, "failed to set add %d", add);
			return TCL_ERROR;
		}

	} else {
		Tcl_AppendResult(interp, "bad option \"", argv[1],
				 "\":  must be data, start, stop, active, "
				 "passcount, params, or fd", (char *) NULL);
		goto error;
	}
	return result;

error:
	return TCL_ERROR;
}

static int MCSCmd(ClientData clientData, Tcl_Interp* interp,
		  int argc, char** argv)
{
	int stat = 0;
	int mode = 0, mod;
	MCS* mcsPtr;
	char* mcsdev = NULL;

	if (argc < 3 || argc > 4) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " commandname device ?mode?\"",
				 (char *) NULL);
		return TCL_ERROR;
	}

	if (argc == 4) {
		if ( Tcl_GetInt(interp, argv[3], &mode) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for mode");
			return TCL_ERROR;
		}
	}
    
	switch(mode) {
	case 0: { mod = O_RDWR; } break;
	case 1: { mod = O_RDONLY; } break;
	default: {
		sprintf(interp->result,
			"wrong mode = %d, should be 0(rdwr) or 1(rdonly)",
			mode );
		return TCL_ERROR;
	}
	}

	mcsPtr = (MCS *) ckalloc(sizeof(MCS));
        if (mcsPtr == NULL) {
		sprintf(interp->result, "Out of memory\n" );
		return TCL_ERROR;
	}
	mcsdev = argv[2];

	/* TCL replaces the 'open'-call  by its own version */
	mcsPtr->mcsfd = open(mcsdev, mod, 0000);
	if ( mcsPtr->mcsfd < 0 ) {
		sprintf(interp->result, "Failed to open MCS-Device\n" );
		goto fail1;
	}

	stat = ioctl(mcsPtr->mcsfd, EGGMCS913_IOCTL_NCHANNELS,
		     &(mcsPtr->nchannels));
	if ( stat != 0) {
		sprintf(interp->result,
			"ioctl: EGGMCS913_IOCTL_NCHANNELS failed\n" );
		goto fail2;
	}

	mcsPtr->dsize = 0;
	mcsPtr->data = NULL;

	mcsPtr->ydata.numValues = 0;
	mcsPtr->ydata.arraySize = 0;
	mcsPtr->ydata.valueArr = NULL;

	mcsPtr->ysize = 0;

	Tcl_CreateCommand(interp, argv[1], MCSObjectCmd,
			  (ClientData) mcsPtr, (Tcl_CmdDeleteProc*) MCSDelete);
	Tcl_AppendResult(interp, argv[1], (char *) NULL);

	return TCL_OK;


fail2:
	close(mcsPtr->mcsfd);
fail1:
	ckfree((char*) mcsPtr);

	return TCL_ERROR;
}

int Mcs_Init(Tcl_Interp *interp)
{
	Tcl_CreateCommand(interp, "mcs", MCSCmd,
			  (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
	return TCL_OK;
}
